/***************************************************************************
*   Copyright (C) 2004-2005 by Thomas Fischer                             *
*   fischer@unix-ag.uni-kl.de                                             *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
#include <qstringlist.h>
#include <qregexp.h>

#include <bibtexentry.h>
#include <bibtexfile.h>
#include <bibtexencoder.h>
#include <bibtexentryfield.h>

#define max(a,b) ((a)>(b)?(a):(b))

BibTeXEntry::BibTeXEntry( const EntryType entryType, const QString &id, QObject *parent, const char *name )
        : BibTeXElement( parent, name ), m_entryType( entryType ), m_id( id )
{
    // TODO
}

BibTeXEntry::BibTeXEntry( BibTeXEntry *other )
{
    m_entryType = other->m_entryType;
    m_id = other->m_id;
    for ( QValueList<BibTeXEntryField*>::iterator it = other->m_fields.begin(); it != other->m_fields.end(); it++ )
    {
        m_fields.append( new BibTeXEntryField( *it ) );
    }
}

BibTeXEntry::~BibTeXEntry()
{
    for ( QValueList<BibTeXEntryField*>::iterator it = m_fields.begin(); it != m_fields.end(); it++ )
    {
        delete( *it );
    }
}

void BibTeXEntry::setEntryType( const EntryType entryType )
{
    m_entryType = entryType;
}

BibTeXEntry::EntryType BibTeXEntry::entryType()
{
    return m_entryType;
}

void BibTeXEntry::setId( const QString& id )
{
    m_id = id;
}

QString BibTeXEntry::id()
{
    return m_id;
}

QString BibTeXEntry::toString( BibTeXFile::Format format, BibTeXFile::Encoding encoding )
{
    switch ( format )
    {
    case BibTeXFile::formatBibTeX:
        {
            QString result( "@" );

            result.append( entryTypeToString( m_entryType ) );

            result.append( " { " );

            result.append( m_id );

            for ( QValueList<BibTeXEntryField*>::iterator it = m_fields.begin(); it != m_fields.end(); it++ )
            {
                QString text;
                if ( encoding == BibTeXFile::encUTF8 )
                    text = ( *it ) ->value();
                else
                    text = BibTeXEncoder::currentBibTeXEncoder() ->encodeSpecialized( ( *it ) ->value(), ( *it ) ->fieldType() );

                BibTeXFile::StringProtection sp = BibTeXFile::spQuote;
                if ( ( *it ) ->isString() )
                {
                    sp = BibTeXFile::spNone;
                }
                else
                {
                    if ( ( ( *it ) ->fieldType() == BibTeXEntryField::ftTitle ) || ( ( *it ) ->fieldType() == BibTeXEntryField::ftBookTitle ) )
                        sp = BibTeXFile::spBoth;
                    else if ( text.contains( '"' ) > 0 )
                        sp = BibTeXFile::spParanthesis;
                }

                result.append( ",\n\t" );
                result.append( ( *it ) ->key() );
                result.append( " = " );
                switch ( sp )
                {
                case BibTeXFile::spNone:
                    break;
                case BibTeXFile::spParanthesis:
                    {
                        int count = 0;
                        for ( unsigned int i = 0; i < text.length(); i++ )
                        {
                            if ( text[ i ] == '}' && ( i == 0 || text[ i - 1 ] != '\\' ) )
                                count++;
                            if ( text[ i ] == '{' && ( i == 0 || text[ i - 1 ] != '\\' ) )
                                count--;
                        }
                        count = max( 0, count ) + 1;

                        for ( int i = 0;i < count; i++ )
                            result.append( "{" );
                    }
                    break;
                case BibTeXFile::spQuote:
                    result.append( "\"" );
                    break;
                case BibTeXFile::spBoth:
                    result.append( "\"{" );
                    break;
                }

                result.append( text );

                switch ( sp )
                {
                case BibTeXFile::spNone:
                    break;
                case BibTeXFile::spParanthesis:
                    {
                        int count = 0;
                        for ( unsigned int i = 0; i < text.length(); i++ )
                        {
                            if ( text[ i ] == '{' && ( i == 0 || text[ i - 1 ] != '\\' ) )
                                count++;
                            if ( text[ i ] == '}' && ( i == 0 || text[ i - 1 ] != '\\' ) )
                                count--;
                        }
                        count = max( 0, count ) + 1;
                        for ( int i = 0;i < count; i++ )
                            result.append( "}" );
                    }
                    break;
                case BibTeXFile::spQuote:
                    result.append( "\"" );
                    break;
                case BibTeXFile::spBoth:
                    result.append( "}\"" );
                    break;
                }
            }

            return result.append( "\n}" );
        }
        break;


    case BibTeXFile::formatXML:
        {
            QString result( " <bibitem type=\"%1\" label=\"%2\">\n" );
            result = result.arg( entryTypeToString( m_entryType ) ).arg( m_id );

            for ( QValueList<BibTeXEntryField*>::iterator it = m_fields.begin(); it != m_fields.end(); it++ )
            {

                switch ( ( *it ) ->fieldType() )
                {
                case BibTeXEntryField::ftUnknown:
                    //TODO
                    break;
                case BibTeXEntryField::ftAuthor:
                    {
                        QStringList authors = QStringList::split( QRegExp( "\\s+(,|and|&)\\s+", false ), ( *it ) ->value() );
                        result.append( "  <authors>\n" );
                        for ( QStringList::iterator authorIt = authors.begin(); authorIt != authors.end(); authorIt++ )
                        {
                            result.append( "   <person>" );
                            result.append( *authorIt );
                            result.append( "</person>\n" );
                        }
                        result.append( "  </authors>\n" );
                    }
                    break;
                case BibTeXEntryField::ftEditor:
                    {
                        QStringList editors = QStringList::split( QRegExp( "\\s+(,|and|&)\\s+", false ), ( *it ) ->value() );
                        result.append( "  <editors>\n" );
                        for ( QStringList::iterator editorIt = editors.begin(); editorIt != editors.end(); editorIt++ )
                        {
                            result.append( "   <person>" );
                            result.append( *editorIt );
                            result.append( "</person>\n" );
                        }
                        result.append( "  </editors>\n" );
                    }
                    break;
                default:
                    {
                        QString entryName = BibTeXEntryField::fieldTypeToString( ( *it ) ->fieldType() );
                        result.append( QString( "  <%1>%2</%3>\n" ).arg( entryName ).arg( ( *it ) ->value() ).arg( entryName ) );
                    }
                    break;
                }

            }

            return result.append( " </bibitem>" );
        }
        break;

    default:
        break;
    }

    return QString::null;
}

bool BibTeXEntry::containsPattern( const QString & pattern, bool caseSensitive )
{
    bool result = FALSE;

    for ( QValueList<BibTeXEntryField*>::iterator it = m_fields.begin(); !result && ( it != m_fields.end() ); it++ )
        result = ( *it ) ->value().contains( pattern, caseSensitive ) > 0;

    return result;
}

bool BibTeXEntry::addField( BibTeXEntryField *field )
{
    m_fields.append( field );
    return TRUE;
}

BibTeXEntryField* BibTeXEntry::getField( const BibTeXEntryField::FieldType fieldType )
{
    BibTeXEntryField * result = NULL;

    for ( QValueList<BibTeXEntryField*>::iterator it = m_fields.begin(); ( it != m_fields.end() ) && ( result == NULL ); it++ )
        if ( ( *it ) ->fieldType() == fieldType ) result = *it;

    return result;
}

BibTeXEntryField* BibTeXEntry::getField( const QString & key )
{
    BibTeXEntryField * result = NULL;

    for ( QValueList<BibTeXEntryField*>::iterator it = m_fields.begin(); ( it != m_fields.end() ) && ( result == NULL ); it++ )
        if ( ( *it ) ->key().lower() == key.lower() )
            result = *it;

    return result;
}

bool BibTeXEntry::deleteField( const QString & key )
{
    for ( QValueList<BibTeXEntryField*>::iterator it = m_fields.begin(); it != m_fields.end(); it++ )
        if ( ( *it ) ->key().lower() == key.lower() )
        {
            delete( *it );
            m_fields.remove( it );
            return TRUE;
        }

    return FALSE;
}

bool BibTeXEntry::deleteField( const BibTeXEntryField::FieldType fieldType )
{
    for ( QValueList<BibTeXEntryField*>::iterator it = m_fields.begin(); it != m_fields.end(); it++ )
        if ( ( *it ) ->fieldType() == fieldType )
        {
            delete( *it );
            m_fields.remove( it );
            return TRUE;
        }

    return FALSE;
}

QValueList<BibTeXEntryField*>::iterator BibTeXEntry::begin()
{
    return m_fields.begin();
}

QValueList<BibTeXEntryField*>::iterator BibTeXEntry::end()
{
    return m_fields.end();
}

int BibTeXEntry::getFieldCount()
{
    return m_fields.count();
}

QString BibTeXEntry::entryTypeToString( const EntryType entryType )
{
    switch ( entryType )
    {
    case etArticle:
        return QString( "Article" );
    case etBook:
        return QString( "Book" );
    case etBooklet:
        return QString( "Booklet" );
    case etCollection:
        return QString( "Collection" );
    case etElectronic:
        return QString( "Electronic (IEEE)" );
    case etInBook:
        return QString( "InBook" );
    case etInCollection:
        return QString( "InCollection" );
    case etInProceedings:
        return QString( "InProceedings" );
    case etManual:
        return QString( "Manual" );
    case etMastersThesis:
        return QString( "MastersThesis" );
    case etMisc:
        return QString( "Misc" );
    case etPhDThesis:
        return QString( "PhDThesis" );
    case etProceedings:
        return QString( "Proceedings" );
    case etTechReport:
        return QString( "TechReport" );
    case etUnpublished:
        return QString( "Unpublished" );
    default:
        return QString( "Unknown" );
    }
}

BibTeXEntry::EntryType BibTeXEntry::entryTypeFromString( const QString &entryTypeString )
{
    if ( entryTypeString.compare( "article" ) == 0 )
        return etArticle;
    else if ( entryTypeString.compare( "book" ) == 0 )
        return etBook;
    else if ( entryTypeString.compare( "booklet" ) == 0 )
        return etBooklet;
    else if ( entryTypeString.compare( "collection" ) == 0 )
        return etCollection;
    else if ( ( entryTypeString.compare( "electronic" ) == 0 ) || ( entryTypeString.compare( "online" ) == 0 ) || ( entryTypeString.compare( "internet" ) == 0 ) || ( entryTypeString.compare( "webpage" ) == 0 ) )
        return etElectronic;
    else if ( entryTypeString.compare( "inbook" ) == 0 )
        return etInBook;
    else if ( entryTypeString.compare( "incollection" ) == 0 )
        return etInCollection;
    else if ( ( entryTypeString.compare( "inproceedings" ) == 0 ) || ( entryTypeString.compare( "conference" ) == 0 ) )
        return etInProceedings;
    else if ( entryTypeString.compare( "manual" ) == 0 )
        return etManual;
    else if ( entryTypeString.compare( "mastersthesis" ) == 0 )
        return etMastersThesis;
    else if ( entryTypeString.compare( "misc" ) == 0 )
        return etMisc;
    else if ( entryTypeString.compare( "phdthesis" ) == 0 )
        return etPhDThesis;
    else if ( entryTypeString.compare( "proceedings" ) == 0 )
        return etProceedings;
    else if ( entryTypeString.compare( "techreport" ) == 0 )
        return etTechReport;
    else if ( entryTypeString.compare( "unpublished" ) == 0 )
        return etUnpublished;
    else
        return etUnknown;
}

BibTeXEntryField::FieldRequireStatus BibTeXEntry::getRequireStatus( BibTeXEntry::EntryType entryType, BibTeXEntryField::FieldType fieldType )
{
    switch ( entryType )
    {
    case etArticle:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftJournal:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftVolume:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNumber:
        case BibTeXEntryField::ftPages:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftISSN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etBook:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftEditor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftPublisher:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftVolume:
        case BibTeXEntryField::ftSeries:
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftEdition:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftISBN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etBooklet:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftTitle:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftHowPublished:
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftYear:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftISBN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etElectronic:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftHowPublished:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftYear:
        case BibTeXEntryField::ftKey:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etInBook:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftEditor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftPages:
        case BibTeXEntryField::ftPublisher:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftVolume:
        case BibTeXEntryField::ftSeries:
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftEdition:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftCrossRef:
        case BibTeXEntryField::ftChapter:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftISBN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etInCollection:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftBookTitle:
        case BibTeXEntryField::ftPublisher:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftEditor:
        case BibTeXEntryField::ftPages:
        case BibTeXEntryField::ftOrganization:
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftType:
        case BibTeXEntryField::ftISBN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etInProceedings:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftPages:
        case BibTeXEntryField::ftBookTitle:
        case BibTeXEntryField::ftEditor:
        case BibTeXEntryField::ftVolume:
        case BibTeXEntryField::ftNumber:
        case BibTeXEntryField::ftSeries:
        case BibTeXEntryField::ftType:
        case BibTeXEntryField::ftChapter:
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftEdition:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftCrossRef:
        case BibTeXEntryField::ftPublisher:
        case BibTeXEntryField::ftISBN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etManual:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftTitle:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftOrganization:
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftEdition:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftYear:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftISBN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etMastersThesis:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftSchool:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etMisc:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftHowPublished:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftYear:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etPhDThesis:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftSchool:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftISBN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etCollection:
    case etProceedings:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftEditor:
        case BibTeXEntryField::ftPublisher:
        case BibTeXEntryField::ftOrganization:
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftSeries:
        case BibTeXEntryField::ftBookTitle:
        case BibTeXEntryField::ftISBN:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etTechReport:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftInstitution:
        case BibTeXEntryField::ftYear:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftType:
        case BibTeXEntryField::ftNumber:
        case BibTeXEntryField::ftAddress:
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftNote:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    case etUnpublished:
        switch ( fieldType )
        {
        case BibTeXEntryField::ftAuthor:
        case BibTeXEntryField::ftTitle:
        case BibTeXEntryField::ftNote:
            return BibTeXEntryField::frsRequired;
        case BibTeXEntryField::ftMonth:
        case BibTeXEntryField::ftYear:
        case BibTeXEntryField::ftKey:
        case BibTeXEntryField::ftURL:
            return BibTeXEntryField::frsOptional;
        default: return BibTeXEntryField::frsIgnored;
        }
    default:
        return BibTeXEntryField::frsOptional;
    }
}

#include "bibtexentry.moc"
