/***************************************************************************
*   Copyright (C) 2004-2005 by Thomas Fischer                             *
*   fischer@unix-ag.uni-kl.de                                             *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include <qprocess.h>
#include <qwaitcondition.h>
#include <qvaluelist.h>
#include <qstring.h>
#include <qapplication.h>
#include <qiodevice.h>

#include <settings.h>
#include <bibtexfile.h>
#include <bibtexentry.h>
#include <bibtexelement.h>
#include "bibtexfileexporter.h"

BibTeXFileExporter::BibTeXFileExporter( BibTeXFile *bibTeXFile ) : QObject(), file( bibTeXFile ), process( NULL )
{
    wc = new QWaitCondition();
}

BibTeXFileExporter::~BibTeXFileExporter()
{
    delete wc;
}

bool BibTeXFileExporter::save( QIODevice *iodevice, BibTeXFile::Format format, BibTeXFile::Encoding encoding )
{
    bool result = false;

    switch ( format )
    {
    case BibTeXFile::formatBibTeX:
        {
            QTextStream stream( iodevice );

            // first, entries like InBook, InCollection and InProceedings have to be written to ensure proper parsing of crossref fields
            for ( BibTeXFile::ElementList::iterator it = file->elements.begin(); it != file->elements.end(); it++ )
            {
                BibTeXElement *element = *it;
                BibTeXEntry *entry = dynamic_cast<BibTeXEntry*>( element );
                if ( entry == NULL || entry->entryType() == BibTeXEntry::etInCollection || entry->entryType() == BibTeXEntry::etInProceedings || entry->entryType() == BibTeXEntry::etInBook )
                    stream << ( ( BibTeXElement* ) * it ) ->toString( BibTeXFile::formatBibTeX, encoding ) << endl << endl;
            }
            // now, all the remaining elements can be written
            for ( BibTeXFile::ElementList::iterator it = file->elements.begin(); it != file->elements.end(); it++ )
            {
                BibTeXElement *element = *it;
                BibTeXEntry *entry = dynamic_cast<BibTeXEntry*>( element );
                if ( entry != NULL && entry->entryType() != BibTeXEntry::etInCollection && entry->entryType() != BibTeXEntry::etInProceedings && entry->entryType() != BibTeXEntry::etInBook )
                    stream << ( ( BibTeXElement* ) * it ) ->toString( BibTeXFile::formatBibTeX, encoding ) << endl << endl;
            }

            result = true;
        }
        break;

    case BibTeXFile::formatPDF:
        {
            QString workingDir = createTempDir();
            result = exportAsPDF( workingDir, iodevice );
            deleteTempDir( workingDir );
        }
        break;

    case BibTeXFile::formatPS:
        {
            QString workingDir = createTempDir();
            result = exportAsPS( workingDir, iodevice );
            deleteTempDir( workingDir );
        }
        break;

    case BibTeXFile::formatXML:
        {
            QTextStream stream( iodevice );
            stream << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
            stream << "<!DOCTYPE biblopgraphy SYSTEM \"bibtexml.dtd\">" << endl;
            stream << "<!--" << endl;
            stream << "  Created with KBibTeX" << endl;
            stream << "  http://www.unix-ag.uni-kl.de/~fischer/kbibtex/" << endl;
            stream << "-->" << endl;
            stream << "<bibliography>" << endl;

            for ( BibTeXFile::ElementList::iterator it = file->elements.begin(); it != file->elements.end(); it++ )
            {
                stream << ( ( BibTeXElement* ) * it ) ->toString( BibTeXFile::formatXML, BibTeXFile::encUTF8 ) << endl;
            }

            stream << "</bibliography>" << endl;

            result = true;
        }
        break;

    default:
        // nothing
        break;
    }

    return result;
}

bool BibTeXFileExporter::exportAsPS( const QDir& tempDir, QIODevice *iodevice )
{
    bool result = false;

    QString LaTeXFilename = QString( tempDir.canonicalPath() ).append( "/kbibtex.tex" );
    QString BibTeXFilename = QString( tempDir.canonicalPath() ).append( "/kbibtex.bib" );
    QString outputFilename = QString( tempDir.canonicalPath() ).append( "/kbibtex.ps" );

    QStringList cmdLines = QStringList::split( '|', "latex kbibtex.tex|bibtex kbibtex|latex kbibtex.tex|latex kbibtex.tex|dvips -o kbibtex.ps kbibtex.dvi" );

    QFile bibtexFile( BibTeXFilename );
    if ( bibtexFile.open( IO_WriteOnly ) )
    {
        result = save( &bibtexFile );
        bibtexFile.close();
    }

    if ( result && writeLatexFile( LaTeXFilename ) && runProcesses( tempDir, cmdLines ) )
    {
        QFile file( outputFilename );
        if ( file.open( IO_ReadOnly ) )
        {
            Q_ULONG buffersize = 0x10000;
            Q_LONG amount = 0;
            char* buffer = new char[ buffersize ];
            do
            {
                result = ( ( amount = file.readBlock( buffer, buffersize ) ) > -1 ) && ( iodevice->writeBlock( buffer, amount ) > -1 );
            }
            while ( result && amount > 0 );

            file.close();
            delete buffer;
        }
        else
            result = false;
    }
    else
        result = false;

    return result;
}

bool BibTeXFileExporter::exportAsPDF( const QDir& tempDir, QIODevice *iodevice )
{
    bool result = false;

    QString LaTeXFilename = QString( tempDir.canonicalPath() ).append( "/kbibtex.tex" );
    QString BibTeXFilename = QString( tempDir.canonicalPath() ).append( "/kbibtex.bib" );
    QString outputFilename = QString( tempDir.canonicalPath() ).append( "/kbibtex.pdf" );

    QStringList cmdLines = QStringList::split( '|', "pdflatex kbibtex.tex|bibtex kbibtex|pdflatex kbibtex.tex|pdflatex kbibtex.tex" );

    QFile bibtexFile( BibTeXFilename );
    if ( bibtexFile.open( IO_WriteOnly ) )
    {
        result = save( &bibtexFile );
        bibtexFile.close();
    }

    if ( result && writeLatexFile( LaTeXFilename ) && runProcesses( tempDir, cmdLines ) )
    {
        QFile file( outputFilename );
        if ( file.open( IO_ReadOnly ) )
        {
            Q_ULONG buffersize = 0x10000;
            Q_LONG amount = 0;
            char* buffer = new char[ buffersize ];
            do
            {
                result = ( ( amount = file.readBlock( buffer, buffersize ) ) > -1 ) && ( iodevice->writeBlock( buffer, amount ) > -1 );
            }
            while ( result && amount > 0 );

            file.close();
            delete buffer;
        }
        else
            result = false;
    }
    else
        result = false;

    return result;
}

bool BibTeXFileExporter::writeLatexFile( const QString &filename )
{
    QFile latexFile( filename );
    if ( latexFile.open( IO_WriteOnly ) )
    {
        QTextStream ts( &latexFile );
        ts << "\\documentclass{article}\n";
        ts << "\\usepackage[" << kbibtexsettings.exportlanguage << "]{babel}\n";
        ts << "\\usepackage{url}\n";
        ts << "\\bibliographystyle{" << kbibtexsettings.exportbibliographystyle << "}\n";
        ts << "\\begin{document}\n";
        ts << "\\nocite{*}\n";
        ts << "\\bibliography{kbibtex}\n";
        ts << "\\end{document}\n";
        latexFile.close();
        return true;
    }
    else
        return false;

}

bool BibTeXFileExporter::runProcess( const QDir& workingDir, const QStringList &args )
{
    bool result = false;

    if ( process == NULL )
    {
        process = new QProcess( );
        connect( process, SIGNAL( processExited() ), this, SLOT( slotProcessExited() ) );
        connect( process, SIGNAL( readyReadStdout() ), this, SLOT( slotReadProcessOutput() ) );
        connect( process, SIGNAL( readyReadStderr() ), this, SLOT( slotReadProcessOutput() ) );
        process->setWorkingDirectory( workingDir );
        process->setArguments( args );

        if ( process->start() )
        {
            qApp->processEvents();
            while ( process->isRunning() )
            {
                wc->wait( 250 );
                qApp->processEvents();
            }

            result = process->normalExit();
        }

        disconnect( process, SIGNAL( readyReadStdout() ), this, SLOT( slotReadProcessOutput() ) );
        disconnect( process, SIGNAL( readyReadStderr() ), this, SLOT( slotReadProcessOutput() ) );
        disconnect( process, SIGNAL( processExited() ), this, SLOT( slotProcessExited() ) );
        delete( process );
        process = NULL;
    }

    return result;
}

bool BibTeXFileExporter::runProcesses( const QDir& workingDir, const QStringList &commandLines )
{
    bool result = true;
    QApplication::setOverrideCursor( Qt::waitCursor );

    int i = 0;
    emit progress( ++i, commandLines.size() + 1 );
    for ( QStringList::ConstIterator it = commandLines.begin(); result && it != commandLines.end(); it++ )
    {
        QStringList args = QStringList::split( ' ', *it );
        result &= runProcess( workingDir, args );
        qApp->processEvents();
        emit progress( ++i, commandLines.size() + 1 );
        qApp->processEvents();
    }

    QApplication::restoreOverrideCursor();
    return result;
}

QString BibTeXFileExporter::createTempDir()
{
    QString result = QString::null;
    QFile *devrandom = new QFile( "/dev/random" );

    if ( devrandom->open( IO_ReadOnly ) )
    {
        Q_UINT32 randomNumber;
        if ( devrandom->readBlock( ( char* ) & randomNumber, sizeof( randomNumber ) ) > 0 )
        {
            randomNumber |= 0x10000000;
            result = QString( "/tmp/bibtex-%1/" ).arg( randomNumber, sizeof( randomNumber ) * 2, 16 );
            qDebug( "Creating temp dir \"%s\"", result.latin1() );
            if ( !QDir().mkdir( result ) )
                result = QString::null;
        }
        devrandom->close();
    }

    delete devrandom;

    return result;
}

void BibTeXFileExporter::deleteTempDir( const QString& directory )
{
    QDir dir = QDir( directory );
    QStringList subDirs = dir.entryList( QDir::Dirs );
    for ( QStringList::Iterator it = subDirs.begin(); it != subDirs.end(); it++ )
    {
        if ( ( QString::compare( *it, "." ) != 0 ) && ( QString::compare( *it, ".." ) != 0 ) )
            deleteTempDir( *it );
    }
    QStringList allEntries = dir.entryList( QDir::All );
    for ( QStringList::Iterator it = allEntries.begin(); it != allEntries.end(); it++ )
        dir.remove( *it );

    QDir().rmdir( directory );
}

void BibTeXFileExporter::slotReadProcessOutput()
{
/*    if ( process )
    {
        while ( process->canReadLineStdout() )
            qDebug( "stdout: %s", process->readLineStdout().latin1() );
        while ( process->canReadLineStderr() )
            qDebug( "stderr: %s", process->readLineStderr().latin1() );
    }*/
}

void BibTeXFileExporter::slotProcessExited()
{
    qDebug( "slotProcessExited\n" );
    wc->wakeAll();
}

void BibTeXFileExporter::cancel()
{
    if ( process != NULL )
    {
        process->tryTerminate();
        process->kill();
        wc->wakeAll();
    }
}

#include "bibtexfileexporter.moc"
